// Copyright (c) 2014 Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of ACG (www.acg.cs.tau.ac.il).
// You can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Author(s): Michal Kleinbort <balasmic@post.tau.ac.il>
//            Oren Salzman <orenzalz@post.tau.ac.il >          

#ifndef SINGLE_GRID_H__RTG
#define SINGLE_GRID_H__RTG

#include <vector>
#include <unordered_set>
#include <map>

#include <boost/unordered_map.hpp>

#include "Random_utils.h" 
#include "Random_matrix_gen.h"
#include "pairs.h"
#include "Point_d.h"
#include "distance_functors.h"

namespace RTG
{
    /// A class used for defining a single grid.
    /*!
        A single transformed grid that defines a partition 
        of the data into cells of a given size.
        Nearby points that lie in the same cell are reported.
    */
    template<typename _PointDT = Point_d_,  
			 typename _SqrdDistanceFuncT = EuclideanSqrdDistanceFunc<_PointDT> >
    class SingleGrid
    {
    public: 
        typedef _SqrdDistanceFuncT SqrdDistanceFunc;
        typedef _PointDT           PointD;
    protected:
	    typedef std::vector<int>    Cell_index; 
	    typedef int                 Compact_cell_index;
		
        /// hash function for cell indices
	    struct Cell_ind_hash: std::unary_function<Cell_index, std::size_t>
	    {
		    std::size_t operator()(Cell_index const& cell) const
		    {
			    std::size_t seed = 0;
			    for(std::size_t i(0); i < cell.size(); ++i)
			    {
				    boost::hash_combine(seed, cell[i]);
			    }

			    return seed;
		    }
	    };


        /// \brief The type of the map between cell index to compact ones
	    typedef boost::unordered_map <Cell_index, //key type
							            Compact_cell_index,    //value type
		                		        Cell_ind_hash> Cell_index_to_compact_cell_index;

        /// \brief The type of the map between compact cell index to vector of integers (the indices of the data points)
	    typedef std::vector< std::vector<int> > Compact_cell_index_to_int_vec;

    public:

        /// \brief constructor of a single grid with rotation
        SingleGrid(  std::vector<double>& shift, double cell_size, double r_sq, const std::vector<PointD>& points,
                        unsigned int dim, boost::shared_ptr<Eigen::MatrixXd> rotation_matrix, SqrdDistanceFunc* pSqrDistFunc)
            :s_(shift), c_(cell_size), r_sq_(r_sq), points_(points), 
	        neighbors_computed_(false), rotation_mat_(rotation_matrix), 
	        with_rotation_(true),
	        pSqrdDistRetriever_(pSqrDistFunc),
	        dim_(dim)
        {
	        init();
            return;
        }
			
        /// \brief constructor of a single grid without rotation
        SingleGrid(  std::vector<double> shift, double cell_size, double r_sq, const std::vector<PointD>& points,
		                unsigned int dim, SqrdDistanceFunc* pSqrDistFunc)
	        :s_(shift), c_(cell_size), r_sq_(r_sq), points_(points), 
	        neighbors_computed_(false), with_rotation_(false), 
	        pSqrdDistRetriever_(pSqrDistFunc),
	        dim_(dim)
        {
	        init();
            return;
        }

        /// \brief distructor
        ~SingleGrid()
        {
	        cell_indices_.clear();
	        cell_index_to_point_indices_.clear();
	        cell_index_to_compact_cell_index_.clear();
            return;
        }
	
        /// \brief find all r-near pairs and update the auxiliary pairs structure
        void find_all_pairs_single(PairsStruct* pairs)
        {
	        for (std::size_t i(0); i < cell_index_to_point_indices_.size(); ++i)
	        {
		        std::vector<int>& cell_points = cell_index_to_point_indices_[i];
		        for (std::size_t j(0); j < cell_points.size(); ++j)
		        {
			        for (std::size_t k(j+1) ; k < cell_points.size(); ++k)
			        {
				        if (pairs->pair_exists(cell_points[j], cell_points[k]))
					        continue;

				        if (are_neighbors(cell_points[j], cell_points[k]))
					        pairs->add_pair(cell_points[j], cell_points[k]);

			        }
		        }
	        }
            return;
        }

	
    private:

        /// \brief initializes the random grid.
        ///        for each data point, locate its cell in the grid
        ///        Store only non-empty grid cells
        void init()
	    {
		    cell_indices_.resize(points_.size());

		    for (std::size_t i(0); i<points_.size(); ++i)
		    {
			    Compact_cell_index cell_index = get_compact_cell_index(points_[i], dim_);
			    cell_indices_[i] = cell_index;
			    cell_index_to_point_indices_[cell_index].push_back(i);
		    }
	    }

        /// \brief get the compact cell index of a given data point p
	    Compact_cell_index get_compact_cell_index(const PointD& p, unsigned int dim)
	    {
		    Cell_index index;
		    //first get the real index
		    if (with_rotation_)
		    {
			    Eigen::VectorXd vec(dim);
			    for (unsigned int i(0); i < dim; ++i)
				    vec(i) = p[i]; //assuming operator [] is implemented in class PointD
    
			    Eigen::VectorXd rotated_vec = (*rotation_mat_)*vec;
    
			    for (unsigned int i(0); i < dim; ++i)
				    index.push_back(get_coord_index(i,rotated_vec)); 
		    }
		    else //(with_rotation == flase)
		    {
			    for (std::size_t i(0); i < dim; ++i)
				    index.push_back(get_coord_index(i,p)); 
		    }

		    //now translate to compact representation
		    typename Cell_index_to_compact_cell_index::iterator 
			    iter = cell_index_to_compact_cell_index_.find(index);

		    Compact_cell_index new_compact_index;
				
		    if (iter == cell_index_to_compact_cell_index_.end())
		    {
			    new_compact_index = cell_index_to_compact_cell_index_.size();
			    cell_index_to_compact_cell_index_[index] = new_compact_index; 
			    cell_index_to_point_indices_.push_back(std::vector<int>());
		    }
		    else
		    {
			    new_compact_index = iter->second;
		    }
		    return new_compact_index;
	    }

        /// \brief get the ith coordinate index
        ///        used for defining the cell index of a given point
        ///        This method is called when random rotation is used
	    int get_coord_index(int i, Eigen::VectorXd & rotated_vec)
	    {
		    double random_shift_coord_i = s_[i];
		    double val_coord_i = rotated_vec(i);
		    double dev = (val_coord_i + random_shift_coord_i) / c_;
		    return my_floor(dev);
	    }

        /// \brief get the ith coordinate index
        ///        used for defining the cell index of a given point
        ///        This method is called when only a random shift is used
        int get_coord_index(int i, const PointD& p)
	    {
		    double random_shift_coord_i = s_[i];
		    double val_coord_i = p[i]; //assuming operator [] is implemented in class PointD
		    double dev = (val_coord_i + random_shift_coord_i) / c_;
		    return my_floor(dev);
	    }

        /// \brief Our floor implementation
        ///        Can be replaced by a more efficient implementation 
        //         of the floor function in the future.
	    int my_floor(double f)
	    {
		    return (int)std::floor(f); 
	    }

        /// \brief Returns true if the ith and jth points in 
        //         the points array are r-nearest-neighbors
	    bool are_neighbors(int i, int j)
	    {
		    return are_neighbors(points_[i], points_[j]);
	    }

	    /// \brief Returns true if the points p and q are r-nearest-neighbors
	    bool are_neighbors(const PointD& p, const PointD& q)
	    {
		    return ((*pSqrdDistRetriever_)(p,q) <= r_sq_);
	    }

    private: 

	    /// \brief The rotation matrix
	    boost::shared_ptr<Eigen::MatrixXd>  rotation_mat_;
	    /// \brief The shift of the grid 
	    std::vector<double>				    s_;			
	    /// \brief The cell size
	    double							    c_;			
	    /// \brief The squared radius
	    double							    r_sq_;		
	    /// \brief Flag indicating whether the pairs were computed
	    bool							    neighbors_computed_;
	    /// \brief Flag indicating whether to use a random rotation 
	    ///  in addition to the random shidt
	    bool							    with_rotation_;
	    /// \brief reference to the vector of data
	    const std::vector<PointD>&			points_;
	    /// \brief map between a cell index and its compact representation
	    Cell_index_to_compact_cell_index    cell_index_to_compact_cell_index_;
	    /// \brief map between point index and its associated compact cell index
	    std::vector<Compact_cell_index>	    cell_indices_; 
	    /// \brief map between a cell and its associated point (insdices)
	    Compact_cell_index_to_int_vec	    cell_index_to_point_indices_;	
	    /// \brief A functor for computing the squared distance
	    SqrdDistanceFunc*					pSqrdDistRetriever_;
	    /// \brief The dimension of every point
	    unsigned int						dim_;

    }; //SingleGrid



} //namespace RTG
#endif //SINGLE_GRID_H__RTG
